local mplib = require ('mplib')
local kpse = require ('kpse')

local format, join = string.format, table.concat
local sprint = tex.sprint
local abs, sqrt, round = math.abs, math.sqrt, math.round

local file = file or {}

function file.replacesuffix(filename, suffix)
    return (filename:gsub("%.[%a%d]+$","")) .. "." .. suffix
end

function file.stripsuffix (filename) 
    return (filename:gsub("%.[%a%d]+$",""))
end


metapost           = metapost or { }

metapost.showlog   = true
metapost.lastlog   = ""
metapost.n         = 0

function metapost.resetlastlog()
    metapost.lastlog = ""
end

local function finder(name, mode, ftype)
    if mode=="w" then
        return name
    else
        return kpse.find_file(name,ftype)
    end
end

metapost.finder = finder

local preamble = [[
input %s ; dump ;
]]

function metapost.make(name, target, version)
    target = file.replacesuffix(target or name, "mem")
    local mpx = mplib.new {
                ini_version = true,
                find_file = finder,
                job_name = file.stripsuffix(target) }
    if mpx then
        local result = mpx:execute(format(preamble,name))
        mpx:finish()
    end
end

function metapost.load(name)
    local mpx = mplib.new {
            ini_version = false,
            mem_name = file.replacesuffix(name,"mem"),
            find_file = finder
        }
    local result
    if not mpx then
        result = { status = 99, error = "out of memory"}
    end
    return mpx, result
end

function metapost.unload(mpx)
    if mpx then
        mpx:finish()
    end
end

function metapost.reporterror(result)
    if not result then
        metapost.report("mp error: no result object returned")
    elseif result.status > 0 then
        local t, e, l = result.term, result.error, result.log
        if t then
            metapost.report("mp terminal: %s",t)
        end
        if e then
            metapost.report("mp error: %s", e)
        end
        if not t and not e and l then
            metapost.lastlog = metapost.lastlog .. "\n " .. l
            metapost.report("mp log: %s",l)
        else
            metapost.report("mp error: unknown, no error, terminal or log messages")
        end
    else
        return false
    end
    return true
end



function metapost.process(mpx, data)
    local converted, result = false, {}
    mpx = metapost.load(mpx)
    if mpx and data then
       result = mpx:execute(data)
       if not result then
          metapost.report("mp error: no result object returned")
       elseif result.status > 0 then
          metapost.report("mp error: %s",(result.term or "no-term") .. "\n" .. (result.error or "no-error"))
       elseif metapost.showlog then
          metapost.lastlog = metapost.lastlog .. "\n" .. result.term
          metapost.report("mp info: %s",result.term or "no-term")
       elseif result.fig then
          converted = metapost.convert(result)
       end
    else
       metapost.report("mp error: mem file not found")
    end
    return converted, result
end

function metapost.report(...)
    print("mplib: " .. format(...))
end


local function getobjects(result,figure,f)
    return figure:objects()
end

function metapost.convert(result, flusher)
    metapost.flush(result, flusher)
    return true -- done
end

function metapost.comment(message)
--    if message then
--        sprint(format("\\MPLIBtoPDF{%%\\space mps graphic %s: %s}", metapost.n, message))
--    end
end

metapost.flushers = { }
metapost.flushers.pdf = { }

function metapost.flushers.pdf.startfigure(n,llx,lly,urx,ury,message)
    metapost.n = metapost.n + 1
    sprint(format("\\startMPLIBtoPDF{%s}{%s}{%s}{%s}",llx,lly,urx,ury))
    if message then metapost.comment(message) end
end

function metapost.flushers.pdf.stopfigure(message)
    if message then metapost.comment(message) end
    sprint("\\stopMPLIBtoPDF")
end

function metapost.flushers.pdf.flushfigure(pdfliterals) -- table
    if #pdfliterals > 0 then
        sprint("\\MPLIBtoPDF{",join(pdfliterals,"\n"),"}")
    end
end

function metapost.flushers.pdf.textfigure(font,size,text,width,height,depth) -- we could save the factor
    text = text:gsub(".","\\hbox{%1}") -- kerning happens in metapost (i have to check if this is true for mplib)
    sprint(format("\\MPLIBtextext{%s}{%s}{%s}{%s}{%s}",font,size,text,0,-( 7200/ 7227)/65536*depth))
end

local bend_tolerance = 131/65536

local rx, sx, sy, ry, tx, ty, divider = 1, 0, 0, 1, 0, 0, 1

local function pen_characteristics(object)
    if mplib.pen_info then
        local t = mplib.pen_info(object)
        rx, ry, sx, sy, tx, ty = t.rx, t.ry, t.sx, t.sy, t.tx, t.ty
        divider = sx*sy - rx*ry
        return not (sx==1 and rx==0 and ry==0 and sy==1 and tx==0 and ty==0), t.width
    else
        rx, sx, sy, ry, tx, ty, divider = 1, 0, 0, 1, 0, 0, 1
        return false, 1
    end
end

local function concat(px, py) -- no tx, ty here
    return (sy*px-ry*py)/divider,(sx*py-rx*px)/divider
end

local function curved(ith,pth)
    local d = pth.left_x - ith.right_x
    if abs(ith.right_x - ith.x_coord - d) <= bend_tolerance and abs(pth.x_coord - pth.left_x - d) <= bend_tolerance then
        d = pth.left_y - ith.right_y
        if abs(ith.right_y - ith.y_coord - d) <= bend_tolerance and abs(pth.y_coord - pth.left_y - d) <= bend_tolerance then
            return false
        end
    end
    return true
end

local function flushnormalpath(path, t, open)
    local pth, ith
    for i=1,#path do
        pth = path[i]
        if not ith then
            t[#t+1] = format("%f %f m",pth.x_coord,pth.y_coord)
        elseif curved(ith,pth) then
            t[#t+1] = format("%f %f %f %f %f %f c",ith.right_x,ith.right_y,pth.left_x,pth.left_y,pth.x_coord,pth.y_coord)
        else
            t[#t+1] = format("%f %f l",pth.x_coord,pth.y_coord)
        end
        ith = pth
    end
    if not open then
        local one = path[1]
        if curved(pth,one) then
            t[#t+1] = format("%f %f %f %f %f %f c",pth.right_x,pth.right_y,one.left_x,one.left_y,one.x_coord,one.y_coord )
        else
            t[#t+1] = format("%f %f l",one.x_coord,one.y_coord)
        end
    elseif #path == 1 then
        -- special case .. draw point
        local one = path[1]
        t[#t+1] = format("%f %f l",one.x_coord,one.y_coord)
    end
    return t
end

local function flushconcatpath(path, t, open)
    t[#t+1] = format("%f %f %f %f %f %f cm", sx, rx, ry, sy, tx ,ty)
    local pth, ith
    for i=1,#path do
        pth = path[i]
        if not ith then
           t[#t+1] = format("%f %f m",concat(pth.x_coord,pth.y_coord))
        elseif curved(ith,pth) then
            local a, b = concat(ith.right_x,ith.right_y)
            local c, d = concat(pth.left_x,pth.left_y)
            t[#t+1] = format("%f %f %f %f %f %f c",a,b,c,d,concat(pth.x_coord, pth.y_coord))
        else
           t[#t+1] = format("%f %f l",concat(pth.x_coord, pth.y_coord))
        end
        ith = pth
    end
    if not open then
        local one = path[1]
        if curved(pth,one) then
            local a, b = concat(pth.right_x,pth.right_y)
            local c, d = concat(one.left_x,one.left_y)
            t[#t+1] = format("%f %f %f %f %f %f c",a,b,c,d,concat(one.x_coord, one.y_coord))
        else
            t[#t+1] = format("%f %f l",concat(one.x_coord,one.y_coord))
        end
    elseif #path == 1 then
        -- special case .. draw point
        local one = path[1]
        t[#t+1] = format("%f %f l",concat(one.x_coord,one.y_coord))
    end
    return t
end

metapost.flushnormalpath = flushnormalpath

metapost.specials = metapost.specials or { }

function metapost.flush(result,flusher)
    if result then
        local figures = result.fig
        if figures then
            flusher = flusher or metapost.flushers.pdf
            local colorconverter = metapost.colorconverter()
            for f=1, #figures do
                local figure = figures[f]
                local objects = getobjects(result,figure,f)
                local fignum = tonumber((figure:filename()):match("([%d]+)$") or figure:charcode() or 0)
                local t = { }
                local miterlimit, linecap, linejoin, dashed = -1, -1, -1, false
                local bbox = figure:boundingbox()
                local llx, lly, urx, ury = bbox[1], bbox[2], bbox[3], bbox[4] -- faster than unpack
                if urx < llx then
                    -- invalid
                    flusher.startfigure(fignum,0,0,0,0,"invalid",figure)
                    flusher.stopfigure()
                else
                    flusher.startfigure(fignum,llx,lly,urx,ury,"begin",figure)
                    t[#t+1] = "q"
                    if objects then
                        for o=1,#objects do
                            local object = objects[o]
                            local objecttype = object.type
                            if objecttype == "start_bounds" or objecttype == "stop_bounds" then
                                -- skip
                            elseif objecttype == "start_clip" then
                                t[#t+1] = "q"
                                flushnormalpath(object.path,t,false)
                                t[#t+1] = "W n"
                            elseif objecttype == "stop_clip" then
                                t[#t+1] = "Q"
                                miterlimit, linecap, linejoin, dashed = -1, -1, -1, false
                            elseif objecttype == "special" then
                                metapost.specials.register(object.prescript)
                            elseif objecttype == "text" then
                                t[#t+1] = "q"
                                local ot = object.transform -- 3,4,5,6,1,2
                                t[#t+1] = format("%f %f %f %f %f %f cm",ot[3],ot[4],ot[5],ot[6],ot[1],ot[2])
                                flusher.flushfigure(t)
                                t = { }
                                flusher.textfigure(object.font,object.dsize,object.text,object.width,object.height,object.depth)
                                t[#t+1] = "Q"
                            else
                                local currentobject = { -- not needed when no extensions
                                    type = object.type,
                                    miterlimit = object.miterlimit,
                                    linejoin = object.linejoin,
                                    linecap = object.linecap,
                                    color = object.color,
                                    dash = object.dash,
                                    path = object.path,
                                    htap = object.htap,
                                    pen = object.pen,
                                    prescript = object.prescript,
                                    postscript = object.postscript,
                                }
                                --
                                local before, inbetween, after = nil, nil, nil
                                --
                                local prescript = currentobject.prescript
                                if prescript then
                                    -- move test to function
                                    local special = metapost.specials[prescript]
                                    if special then
                                        currentobject, before, inbetween, after = special(currentobject.postscript,currentobject,t,flusher)
                                        objecttype = currentobject.type
                                    end
                                end
                                --
                                local cs, cr = currentobject.color, nil
                                if cs and #cs > 0 then
                                    t[#t+1], cr = colorconverter(cs)
                                end
                                --
                                if before then currentobject, t = before() end
                                local ml = currentobject.miterlimit
                                if ml and ml ~= miterlimit then
                                    miterlimit = ml
                                    t[#t+1] = format("%f M",ml)
                                end
                                local lj = currentobject.linejoin
                                if lj and lj ~= linejoin then
                                    linejoin = lj
                                    t[#t+1] = format("%i j",lj)
                                end
                                local lc = currentobject.linecap
                                if lc and lc ~= linecap then
                                    linecap = lc
                                    t[#t+1] = format("%i J",lc)
                                end
                                local dl = currentobject.dash
                                if dl then
                                    local d = format("[%s] %i d",join(dl.dashes or {}," "),dl.offset)
                                    if d ~= dashed then
                                        dashed = d
                                        t[#t+1] = dashed
                                    end
                                elseif dashed then
                                   t[#t+1] = "[] 0 d"
                                   dashed = false
                                end
                                if inbetween then currentobject, t = inbetween() end
                                local path = currentobject.path
                                local transformed, penwidth = false, 1
                                local open = path and path[1].left_type and path[#path].right_type
                                local pen = currentobject.pen
                                if pen then
								   if pen.type == 'elliptical' then
                                        transformed, penwidth = pen_characteristics(object) -- boolean, value
                                        t[#t+1] = format("%f w",penwidth)
                                        if objecttype == 'fill' then
                                            objecttype = 'both'
                                        end
                                   else -- calculated by mplib itself
                                        objecttype = 'fill'
                                   end
                                end
                                if transformed then
                                    t[#t+1] = "q"
                                end
                                if path then
                                    if transformed then
                                        flushconcatpath(path,t,open)
                                    else
                                        flushnormalpath(path,t,open)
                                    end
                                    if objecttype == "fill" then
                                        t[#t+1] = "h f"
                                    elseif objecttype == "outline" then
                                        t[#t+1] = (open and "S") or "h S"
                                    elseif objecttype == "both" then
                                        t[#t+1] = "h B"
                                    end
                                end
                                if transformed then
                                    t[#t+1] = "Q"
                                end
                                local path = currentobject.htap
                                if path then
                                    if transformed then
                                        t[#t+1] = "q"
                                    end
                                    if transformed then
                                        flushconcatpath(path,t,open)
                                    else
                                        flushnormalpath(path,t,open)
                                    end
                                    if objecttype == "fill" then
                                        t[#t+1] = "h f"
                                    elseif objecttype == "outline" then
                                        t[#t+1] = (open and "S") or "h S"
                                    elseif objecttype == "both" then
                                        t[#t+1] = "h B"
                                    end
                                    if transformed then
                                        t[#t+1] = "Q"
                                    end
                                end
                                if cr then
                                    t[#t+1] = cr
                                end
                                if after then currentobject, t = after() end
                            end
                       end
                    end
                    t[#t+1] = "Q"
                    flusher.flushfigure(t)
                    flusher.stopfigure("end")
                end
            end
        end
    end
end

function metapost.parse(result)
    if result then
        local figures = result.fig
        if figures then
            for f=1, #figures do
                local figure = figures[f]
                local objects = getobjects(result,figure,f)
                if objects then
                    for o=1,#objects do
                        local object = objects[o]
                        if object.type == "outline" then
                            local prescript = object.prescript
                            if prescript then
                                local special = metapost.specials[prescript]
                                if special then
                                    special(object.postscript,object)
                                end
                            end
                       end
                    end
                end
            end
        end
    end
end

function metapost.colorconverter()
    return function(cr)
        local n = #cr
        if n == 4 then
            local c, m, y, k = cr[1], cr[2], cr[3], cr[4]
            return format("%.3f %.3f %.3f %.3f k %.3f %.3f %.3f %.3f K",c,m,y,k,c,m,y,k), "0 g 0 G"
        elseif n == 3 then
            local r, g, b = cr[1], cr[2], cr[3]
            return format("%.3f %.3f %.3f rg %.3f %.3f %.3f RG",r,g,b,r,g,b), "0 g 0 G"
        else
            local s = cr[1]
            return format("%.3f g %.3f G",s,s), "0 g 0 G"
        end
    end
end

